# Claudio Perez
# Summer 2024
import base64
import textwrap
from pathlib import Path

class Viewer:
    """
    A class to represent a 3D model viewer.

    Methods:
    --------
    __init__(self, viewer=None, path=None, data=None):
        Initializes the Viewer with optional viewer type, file path, or binary data.

    """
    def __init__(self, viewer=None, path=None, data=None):
        if data is not None:
            data64 = base64.b64encode(data).decode('utf-8')
            self._glbsrc=f"data:model/gltf-binary;base64,{data64}"
        else:
            self._glbsrc = path 

        self._viewer = viewer if viewer is not None else "mv"

    def get_html(self):
        if self._viewer == "babylon":
            with open(Path(__file__).parents[0]/"babylon.html", "r") as f:
                return f.read()

        if self._viewer == "three-170":
            with open(Path(__file__).parents[0]/"three-170.html", "r") as f:
                return f.read()

        if self._viewer == "three-160":
            with open(Path(__file__).parents[0]/"gltf.html", "r") as f:
                return f.read()

        elif self._viewer == "three-130":
            with open(Path(__file__).parents[0]/"index.html", "r") as f:
                return f.read()

        elif self._viewer == "mv":
            return _model_viewer(self._glbsrc, control=False)


def _model_viewer(source, control=False):
      library = '<script type="module" src="https://ajax.googleapis.com/ajax/libs/model-viewer/4.0.0/model-viewer.min.js"></script>'

      try:
          with open(Path(__file__).parents[0]/"controls.css", "r") as f:
              control_style = f"<style>{f.read()}</style>"

          with open(Path(__file__).parents[0]/"controls.js", "r") as f:
              control_code = f"<script>{f.read()}</script>"
      except:
          control_code = ""
          control_style = ""

      control_html = """
        <div class="controls">
          <!-- <button id="step-backward">Step Left</button> -->
          <button id="toggle-animation">Pause</button>
          <!-- <button id="step-forward">Step Right</button> -->
        </div>
      """


      foot = "</body></html>"
      head = f"""
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
          "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
        <head>
          <meta charset="utf-8">
          <meta name="viewport" content="width=device-width, initial-scale=1">
          <title>veux</title>
          {library}
          {control_style}
        </head>
        <body>
      """

      viewer = f"""
          <model-viewer id="veux-viewer"
                        alt="rendering"
                        src="{source}"
                        autoplay
                        style="width: 100%; height: 500px;"
                        max-pixel-ratio="2"
                        shadow-intensity="1"
                        environment-image="/black_ground.hdr"
                        environment-image="neutral"
                        shadow-light="10000 10000 10000"
                        exposure="0.8"
                        camera-controls
                        min-camera-orbit="auto auto 0m"
                        touch-action="pan-y">
          </model-viewer>
      """
      return textwrap.dedent(f"""
            {head}
            {viewer}
            {control_html if control else ""}
            {control_code if control else ""}
            {foot}
      """)



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Main command line interface
#
import sys

from veux import render
from veux.parser import parse_args, NAME
from veux.errors import RenderError
import veux.server


def main(argv):

    try:
        config = parse_args(argv)

        if config is None:
            sys.exit()

        artist = render(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])
            return

        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            import veux.server
            viewer = config["viewer_config"].get("name", None)
            port = config["server_config"].get("port", None)
            server = veux.server.Server(glb=artist.canvas.to_glb(),
                                        viewer=viewer)
            server.run(port=port)

        elif hasattr(artist.canvas, "to_html"):
            import veux.server
            port = config["server_config"].get("port", None)
            server = veux.server.Server(html=artist.canvas.to_html())
            server.run(port=port)

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(f"ERROR - {e}", file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=NAME), file=sys.stderr)
        sys.exit(-1)


if __name__ == "__main__":
    main(sys.argv)




"""
"""
import numpy as np


def _is_clockwise(polygon):
    """
    Check if a polygon is defined in clockwise order.
    """
    return np.sum((polygon[1:, 0] - polygon[:-1, 0]) * (polygon[1:, 1] + polygon[:-1, 1])) > 0

def _is_convex(a, b, c):
    """
    Check if three points make a convex corner.
    """
    ax, ay = a
    bx, by = b
    cx, cy = c
    return (bx - ax) * (cy - ay) - (by - ay) * (cx - ax) > 0

def _is_ear(polygon, i):
    """
    Check if a vertex at index `i` forms an ear in the polygon.
    """
    prev_idx = (i - 1) % len(polygon)
    next_idx = (i + 1) % len(polygon)
    p_prev = polygon[prev_idx]
    p_curr = polygon[i]
    p_next = polygon[next_idx]

    # Triangle formed by the current vertex and its neighbors
    ear_triangle = np.array([p_prev, p_curr, p_next])

    # Check if the triangle is convex
    if not _is_convex(p_prev, p_curr, p_next):
        return False

    # Check if any other vertex lies inside the triangle
    for j, point in enumerate(polygon):
        if j not in [prev_idx, i, next_idx]:
            if _point_in_triangle(point, ear_triangle):
                return False

    return True

def _point_in_triangle(p, triangle):
    """
    Check if a point `p` is inside the triangle defined by three vertices.
    """
    def sign(a, b, c):
        return (a[0] - c[0]) * (b[1] - c[1]) - (b[0] - c[0]) * (a[1] - c[1])

    a, b, c = triangle
    b1 = sign(p, a, b) < 0
    b2 = sign(p, b, c) < 0
    b3 = sign(p, c, a) < 0

    return b1 == b2 == b3


def earcut2(vertices):
    """
    Triangulate a concave polygon using the ear clipping algorithm.

    Parameters:
        polygon_vertices (array-like): List or numpy array of (x, y) coordinates defining the polygon.

    Returns:
        list of tuple: Indices of triangles as (i, j, k) referencing the input vertices.
    """
    # Coerce input to numpy array
    vertices = np.asarray(vertices, dtype=float)

    # Remove duplicate last point if it's the same as the first
    if np.array_equal(vertices[0], vertices[-1]):
        vertices = vertices[:-1]

    if len(vertices) < 3:
        raise ValueError("A polygon must have at least 3 vertices to triangulate.")

    # Ensure the polygon is in counter-clockwise order
    if _is_clockwise(vertices):
        vertices = vertices[::-1]

    # Copy vertices and initialize the triangle list
    polygon = vertices.tolist()
    triangles = []

    # Loop until the polygon is reduced to three vertices
    while len(polygon) > 3:
        for i in range(len(polygon)):
            if _is_ear(vertices, i):
                # Add the ear triangle to the result
                prev_idx = (i - 1) % len(polygon)
                next_idx = (i + 1) % len(polygon)
                triangles.append((prev_idx, i, next_idx))

                # Remove the ear vertex from the polygon
                del polygon[i]
                break

    # Add the last remaining triangle
    triangles.append((0, 1, 2))

    return triangles


def earcut3(vertices):
    """
    Triangulate a simple polygon using the ear clipping algorithm.

    Parameters:
    polycoord (list or ndarray): A list or array of [x, y] coordinates of the polygon vertices in order.

    Returns:
    triangles (ndarray): An (n, 3) array of indices into polycoord, each row representing a triangle.
    """

    def is_convex(a, b, c):
        """Check if the angle formed by points a, b, c is convex."""
        return (b[0] - a[0])*(c[1] - a[1]) - (b[1] - a[1])*(c[0] - a[0]) < 0

    def point_in_triangle(a, b, c, p):
        """Check if point p is inside triangle abc."""
        # Barycentric coordinate method
        detT = (b[1] - c[1])*(a[0] - c[0]) + (c[0] - b[0])*(a[1] - c[1])
        if detT == 0:
            return False
        l1 = ((b[1] - c[1])*(p[0] - c[0]) + (c[0] - b[0])*(p[1] - c[1])) / detT
        l2 = ((c[1] - a[1])*(p[0] - c[0]) + (a[0] - c[0])*(p[1] - c[1])) / detT
        l3 = 1 - l1 - l2
        return (0 < l1 < 1) and (0 < l2 < 1) and (0 < l3 < 1)

    # Convert to numpy array
    vertices = np.asarray(vertices)

    # Remove duplicate last point if it's the same as the first
    if np.array_equal(vertices[0], vertices[-1]):
        vertices = vertices[:-1]

    n = len(vertices)
    if n < 3:
        raise ValueError("A polygon must have at least 3 vertices.")

    # Ensure the polygon is counter-clockwise
    area = 0.0
    for i in range(n):
        j = (i + 1) % n
        area += (vertices[j][0] - vertices[i][0]) * (vertices[j][1] + vertices[i][1])
    if area > 0:
        vertices = vertices[::-1]

    indices = list(range(len(vertices)))
    triangles = []

    while len(indices) > 3:
        ear_found = False
        for i in range(len(indices)):
            prev_idx = indices[i - 1]
            curr_idx = indices[i]
            next_idx = indices[(i + 1) % len(indices)]

            a = vertices[prev_idx]
            b = vertices[curr_idx]
            c = vertices[next_idx]

            if is_convex(a, b, c):
                # Check if any other vertex is inside the triangle abc
                ear = True
                for idx in indices:
                    if idx in (prev_idx, curr_idx, next_idx):
                        continue
                    p = vertices[idx]
                    if point_in_triangle(a, b, c, p):
                        ear = False
                        break
                if ear:
                    triangles.append([prev_idx, curr_idx, next_idx])
                    del indices[i]
                    ear_found = True
                    break

        if not ear_found:
            raise ValueError("No ear found. The polygon might be not simple or is degenerate.")

    # Add the last remaining triangle
    triangles.append([indices[0], indices[1], indices[2]])

    return np.array(triangles)


def earcut4(polygon, tolerance=1e-6):
    """
    Improved Ear Clipping algorithm with pre-sorting and tolerance for degenerate cases
    """
    from shapely import Polygon
    if isinstance(polygon, Polygon):
        vertices = list(polygon.exterior.coords)[:-1]  # Ignore the last point because it's a repetition of the first
    else:
        vertices = polygon
        polygon = Polygon(vertices)

    if vertices[-1][0] == vertices[0][0] and vertices[-1][1] == vertices[0][1]:
        vertices = vertices[:-1]
    triangles = []

    # Sort vertices by x-coordinate
    vertices = sorted(vertices, key=lambda vertex: vertex[0])

    while len(vertices) > 3:
        for i in range(len(vertices)):
            a, b, c = vertices[i - 1], vertices[i], vertices[(i + 1) % len(vertices)]
            triangle = Polygon([a, b, c])

            # Adjusted checks for valid "ear"
            if triangle.is_valid and triangle.area > tolerance and triangle.contains(polygon):
                triangles.append(triangle)
                vertices.pop(i)
                break
        else:
            raise Exception("No valid ear found")

    if len(vertices) == 3:
        triangles.append(Polygon(vertices))

    return triangles




"""
TODO: Add morph targets
"""
import math
import struct
import numpy as np
import pygltflib

from veux.canvas.gltf import GltfLibCanvas


def _append_index(lst, item):
    lst.append(item)
    return len(lst) - 1


class GltfLibAnimation01:
    """
    A utility class that manages animations for a glTF scene.
    It owns a GltfLibCanvas instance which holds geometry, nodes, etc.
    """

    def __init__(self, canvas=None, name="MyAnimation"):
        if canvas is None:
            canvas = GltfLibCanvas(config=None)

        self.canvas = canvas

        gltf = canvas.gltf

        # Build the Animation object
        anim = pygltflib.Animation(
            samplers=[],
            channels=[],
            name=name
        )

        # Append to glTF's animations
        if gltf.animations is None:
            gltf.animations = []
        gltf.animations.append(anim)
        self.anim = anim


    def set_node_states(self, node_index, times, rotations, path="rotation"):
        """
        Attach a keyframed quaternion-rotation animation to a node in the scene.

        :param node_index: index of the node in self.canvas.gltf.nodes
        :param times:      list/array of timestamps
        :param quaternions: list/array of (x, y, z, w) quaternions
        :param path:       "rotation" (for rotating the node), or "translation", etc.
        """
        gltf = self.canvas.gltf
        canvas = self.canvas

        # Convert times/quaternions to binary
        time_bytes = b"".join([struct.pack("<f", t) for t in times])
        quat_bytes = b"".join([struct.pack("<4f", *q) for q in rotations])

        # Create Accessors for time & states
        time_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
                bufferView=canvas._push_data(time_bytes, target=None),
                byteOffset=0,
                componentType=pygltflib.FLOAT,
                count=len(times),
                type=pygltflib.SCALAR,
                min=[times[ 0]] if times else [0],
                max=[times[-1]] if times else [0]
            )
        )

        state_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
            bufferView=canvas._push_data(quat_bytes, target=None),
            byteOffset=0,
            componentType=pygltflib.FLOAT,
            count=len(rotations),
            type=pygltflib.VEC4
        ))


        # Create AnimationSampler & AnimationChannel
        sampler = pygltflib.AnimationSampler(
            input=time_accessor,
            output=state_accessor,
            interpolation="LINEAR"
        )
        channel = pygltflib.AnimationChannel(
            sampler=0,  # within this animation's local array
            target=pygltflib.AnimationChannelTarget(
                node=node_index,
                path=path
            )
        )

        self.anim.samplers.append(sampler)
        self.anim.channels.append(channel)


    def save(self, filename="scene.glb"):
        """Delegate to the canvas to write out the final GLB."""
        self.canvas.write(filename)


def _create_rotations(
    period = 2.0,
    num_samples = 10,
    amplitude_degs_z = 45,
    amplitude_degs_x = 10):

    def quaternion_multiply(q1, q2):
        """
        Standard quaternion multiply: q1 * q2
        q1, q2 are (x, y, z, w)
        """
        x1, y1, z1, w1 = q1
        x2, y2, z2, w2 = q2
        x = w1*x2 + x1*w2 + y1*z2 - z1*y2
        y = w1*y2 + y1*w2 + z1*x2 - x1*z2
        z = w1*z2 + z1*w2 + x1*y2 - y1*x2
        w = w1*w2 - x1*x2 - y1*y2 - z1*z2
        return (x, y, z, w)


    amp_z = math.radians(amplitude_degs_z)
    amp_x = math.radians(amplitude_degs_x)
    omega = 2 * math.pi / period

    times = []
    quaternions = []

    for i in range(num_samples):
        t = i * (period / (num_samples - 1)) if (num_samples > 1) else 0
        angle_z = amp_z * math.cos(omega * t)
        angle_x = amp_x * math.sin(omega * t)

        # Convert to quaternions
        half_z = angle_z / 2
        half_x = angle_x / 2

        qZ = (0.0, 0.0, math.sin(half_z), math.cos(half_z))
        qX = (math.sin(half_x), 0.0, 0.0, math.cos(half_x))

        qTotal = quaternion_multiply(qZ, qX)

        times.append(t)
        quaternions.append(qTotal)
    return times, quaternions 


def _test_pendulum():
    # 1) Create glTF animation helper
    anim = GltfLibAnimation01()

    # 2) Make pendulum geometry
    #    Define two vertices for a line: pivot at (0,0,0), tip at (0, -L, 0)
    L = 2.0
    anim.canvas.plot_lines(
        vertices=np.array([[0,  0, 0], [0, -L, 0]],dtype=float),
        indices=[[0, 1]],
    )

    # The newly added line is in a new node at the end of gltf.nodes
    node = len(anim.canvas.gltf.nodes) - 1

    times, rotations = _create_rotations()

    # 4) Add an animation to rotate our pendulum node
    anim.set_node_states(node, times, rotations)

    # 5) Save everything to disk
    anim.save("pendulum.glb")


def _test_pendulum02():
    from veux.frame.skins import VeuxAnimation
    # 1) Create glTF animation helper
    anim = VeuxAnimation()

    # 2) Make pendulum geometry
    #    Define two vertices for a line: pivot at (0,0,0), tip at (0, -L, 0)
    L = 2.0
    canvas = GltfLibCanvas()
    lines = canvas.plot_lines(
        vertices=np.array([[0,  0, 0], [0, -L, 0]],dtype=float),
        indices=[[0, 1]],
    )

    # The newly added line is in a new node at the end of gltf.nodes
    node = lines[0].id

    for time, rotation in zip(*_create_rotations()):
        anim.set_node_rotation(node, rotation, time=time)

    anim.apply(canvas)

    # 5) Save everything to disk
    canvas.write("pendulum02.glb")


if __name__ == "__main__":
    _test_pendulum()
    _test_pendulum02()




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
#
import sys
import warnings
import numpy as np
from scipy.spatial.transform import Rotation

import shps.curve

import veux
import veux.frame
from veux.utility.earcut import earcut, flatten as flatten_earcut
from veux.model import read_model
from veux.errors import RenderError
from veux.config import MeshStyle
from shps.frame.extrude import FrameMesh
from dataclasses import dataclass

@dataclass
class ExtrusionCollection:
    triang: list
    coords: list
    caps: list
    no_outline: set
    in_outline: set

def add_extrusion(extr, e, x, R, I, caps=None):

    ring_ranges = extr.ring_ranges()

    p = extr.vertices()
    indices = extr.triangles()

    if len(indices) == 0 or len(ring_ranges) == 0:
        return 0

    e.triang.extend([I + T for T in indices])

    for (j, start_idx, end_idx) in ring_ranges:
        for i in range(start_idx, end_idx):
            e.coords.append(x[j] + R[j] @ p[i])

    if caps:
        nen = len(x)
        noe = ring_ranges[0][-1]
        caps[0].append(            I+np.arange(noe))
        caps[1].append((nen-1)*noe+I+np.arange(noe))

    return len(indices)


def draw_extrusions3(model, canvas, state=None, config=None):
    if config is None:
        config = {"style": MeshStyle(color="gray")}

    scale = config.get("scale", 1.0)

    # 1) Build local geometry
    #----------------------------------------------------------
    I = 0
    caps = []
    e = ExtrusionCollection([], [], [], set(), set())
    for tag in model.iter_cell_tags():

        X_ref = np.array(model.cell_position(tag))
        nen = len(X_ref)

        if state is not None:
            u = state.cell_array(tag, state.position)
            x = shps.curve.displace(X_ref, u, nen)
            R = state.cell_array(tag, state.rotation)
        else:
            x = X_ref
            R = [model.frame_orientation(tag).T]*nen

        sections = [model.frame_section(tag, i) for i in range(len(x))]
        si = sections[0]
        if sections[0] is None or sections[-1] is None:
            continue

        icap, jcap = [], []
        #
        # Exterior
        #
        extr = FrameMesh(len(x),
                        [s.exterior() for s in sections],
                        scale=scale,
                        do_end_caps=False)

        ne = add_extrusion(extr, e, x, R, I, [icap, jcap])

        if len(si.exterior()) > 25:
            for i in range(ne):
                e.no_outline.add(I+i)

        #
        # Interior
        #
        ni = 0
        for i in range(len(si.interior())):
            if si.interior()[i] is None or len(si.interior()[i]) == 0:
                continue

            extr = FrameMesh(len(x),
                    [s.interior()[i] for s in sections],
                    scale=scale,
                    do_end_caps=False)
        
            nij = add_extrusion(extr, e, x, R, I+ne+ni, [icap, jcap])
            for i in range(nij):
                # e.no_outline.add(I+ne+ni+i)
                e.in_outline.add(I+ne+ni+i)
            ni += nij

        I += ni + ne

        #
        # Caps
        #
        try:
            face = si.triangles()

        except Exception as ex:
            warnings.warn(f"Earcut failed with message: {ex}")
            continue

        iicap = [ i for j in icap for i in j ]
        ijcap = [ i for j in jcap for i in j ]
        caps.extend([
            [iicap[i] for i in face],
            [ijcap[i] for i in face]
        ])


    # Draw mesh
    mesh = canvas.plot_mesh(e.coords, 
                     [list(reversed(face)) for face in e.triang], 
                     style=config["style"])

    # Draw caps
    if len(caps) > 0:
        for cap in caps:
            try:
                canvas.plot_mesh(mesh.vertices, cap, style=config["style"])
            except Exception as ex:
                print(ex)


    # Draw outlines
    if "outline" not in config:
        return

    triang = e.triang
    nan = np.array([0,0,0], dtype=float)*np.nan
    IDX = np.array(((0,2),(0,1)))
    coords = np.array(e.coords)

    # canvas.plot_lines( np.array([
    #         coords[idx] if (j+1)%3 else nan
    #         for j,idx in enumerate(np.array(triang).reshape(-1))
    # ]))
    if "tran" in config["outline"]:
        tri_points = np.array([
            coords[idx]  if (j+1)%3 else nan
            for j,idx in enumerate(np.array(triang).reshape(-1))
        ])

    elif "long" in config["outline"]:
        tri_points = np.array([
            coords[i]  if j%2 else nan
            for j,idx in enumerate(np.array(triang)) for i in idx[IDX[j%2]] if j not in e.no_outline
        ])
    else:
        return

    canvas.plot_lines(tri_points,
                      style=config["line_style"])


def draw_extrusions(model, canvas, state=None, config=None):
    ndm = 3

    coords = [] # Global mesh coordinates
    triang = []
    caps   = []
    locoor = [] # Local mesh coordinates, used for textures

    if config is None:
        config = {
                "style": MeshStyle(color="gray")
        }

    scale_section = config["scale"]


    I = 0
    # Track outlines with excessive edges (eg, circles) to later avoid showing
    # their edges
    no_outline = set()
    for tag in model.iter_cell_tags():

        section = model.frame_section(tag)
        if section is None:
            continue

        outline_scale = scale_section

        nen  = len(model.cell_nodes(tag))

        Xi = model.cell_position(tag)
        if state is not None:
            glob_displ = state.cell_array(tag, state.position)
            X = shps.curve.displace(Xi, glob_displ, nen).T
            R = state.cell_array(tag, state.rotation)
        else:
            outline_scale *= 0.99
            X = np.array(Xi)
            R = [model.frame_orientation(tag).T]*nen


        noe = len(section.exterior())
        try:
            face_i = model.frame_section(tag, 0).exterior()[:,1:]
            face_j = model.frame_section(tag, 1).exterior()[:,1:]
            caps.append(I+np.array(earcut(face_i)))
            caps.append(I+(nen-1)*noe + np.array(earcut(face_j)))
        except Exception as e:
            warnings.warn(f"Earcut failed with message: {e}")

        # Loop over sample points along element length to assemble
        # `coord` and `triang` arrays
        for j in range(nen):
            section = model.frame_section(tag, j) # TODO: Pass float between 0 and 1 instead of j
            outline = section.exterior().copy()
            outline[:,1:] *= outline_scale
            # Loop over section edges
            for k,edge in enumerate(outline):
                # Append rotated section coordinates to list of coordinates
                coords.append(X[j, :] + R[j]@edge)
                locoor.append([ (j+0)/nen+0.1,  0.1+(k+0)/(noe+0) ])

                if j == 0:
                    # Skip the first section
                    continue

                elif k < noe-1:
                    triang.extend([
                        [I+    noe*j + k,   I+    noe*j + k + 1,    I+noe*(j-1) + k],
                        [I+noe*j + k + 1,   I+noe*(j-1) + k + 1,    I+noe*(j-1) + k]
                    ])
                else:
                    # elif j < N-1:
                    triang.extend([
                        [I+    noe*j + k,    I + noe*j , I+noe*(j-1) + k],
                        [      I + noe*j, I + noe*(j-1), I+noe*(j-1) + k]
                    ])

                if len(outline) > 25:
                    no_outline.add(len(triang)-1)
                    no_outline.add(len(triang)-2)

        I += nen*noe

    triang = [list(reversed(i)) for i in triang]

    if len(triang) == 0:
        return

    mesh = canvas.plot_mesh(coords, triang, local_coords=locoor, style=config["style"])

    if len(caps) > 0:
        for cap in caps:
            try:
                canvas.plot_mesh(mesh.vertices, cap, style=config["style"])
            except:
                pass


    IDX = np.array((
        (0, 2),
        (0, 1)
    ))

    triang = [list(reversed(i)) for i in triang]

    nan = np.zeros(ndm)*np.nan
    coords = np.array(coords)
    if "tran" in config["outline"]:
        tri_points = np.array([
            coords[idx]  if (j+1)%3 else nan
            for j,idx in enumerate(np.array(triang).reshape(-1))
        ])
    elif "long" in config["outline"]:
        tri_points = np.array([
            coords[i]  if j%2 else nan
            for j,idx in enumerate(np.array(triang)) for i in idx[IDX[j%2]] if j not in no_outline
        ])
    else:
        return

    canvas.plot_lines(tri_points,
                      style=config["line_style"]
    )

def draw_extrusions(model, canvas, state=None, config=None):
    ndm = 3

    coords = [] # Global mesh coordinates
    triang = []
    caps   = []
    locoor = [] # Local mesh coordinates, used for textures

    if config is None:
        config = {
                "style": MeshStyle(color="gray")
        }

    scale_section = config["scale"]


    I = 0
    # Track outlines with excessive edges (eg, circles) to later avoid showing
    # their edges
    no_outline = set()
    for tag in model.iter_cell_tags():

        section = model.frame_section(tag)
        if section is None:
            continue

        outline_scale = scale_section

        nen  = len(model.cell_nodes(tag))

        Xi = model.cell_position(tag)
        if state is not None:
            glob_displ = state.cell_array(tag, state.position)
            X = shps.curve.displace(Xi, glob_displ, nen).T
            R = state.cell_array(tag, state.rotation)
        else:
            outline_scale *= 0.99
            X = np.array(Xi)
            R = [model.frame_orientation(tag).T]*nen


        noe = len(section.exterior())
        try:
            face_i = model.frame_section(tag, 0).exterior()[:,1:]
            face_j = model.frame_section(tag, 1).exterior()[:,1:]
            caps.append(I+np.array(earcut(face_i)))
            caps.append(I+(nen-1)*noe + np.array(earcut(face_j)))
        except Exception as e:
            warnings.warn(f"Earcut failed with message: {e}")

        # Loop over sample points along element length to assemble
        # `coord` and `triang` arrays
        for j in range(nen):
            section = model.frame_section(tag, j) # TODO: Pass float between 0 and 1 instead of j
            outline = section.exterior().copy()
            outline[:,1:] *= outline_scale
            # Loop over section edges
            for k,edge in enumerate(outline):
                # Append rotated section coordinates to list of coordinates
                coords.append(X[j, :] + R[j]@edge)
                locoor.append([ (j+0)/nen+0.1,  0.1+(k+0)/(noe+0) ])

                if j == 0:
                    # Skip the first section
                    continue

                elif k < noe-1:
                    triang.extend([
                        [I+    noe*j + k,   I+    noe*j + k + 1,    I+noe*(j-1) + k],
                        [I+noe*j + k + 1,   I+noe*(j-1) + k + 1,    I+noe*(j-1) + k]
                    ])
                else:
                    # elif j < N-1:
                    triang.extend([
                        [I+    noe*j + k,    I + noe*j , I+noe*(j-1) + k],
                        [      I + noe*j, I + noe*(j-1), I+noe*(j-1) + k]
                    ])

                if len(outline) > 25:
                    no_outline.add(len(triang)-1)
                    no_outline.add(len(triang)-2)

        I += nen*noe

    triang = [list(reversed(i)) for i in triang]

    if len(triang) == 0:
        return

    mesh = canvas.plot_mesh(coords, triang, local_coords=locoor, style=config["style"])

    if len(caps) > 0:
        for cap in caps:
            try:
                canvas.plot_mesh(mesh.vertices, cap, style=config["style"])
            except:
                pass


    IDX = np.array((
        (0, 2),
        (0, 1)
    ))

    triang = [list(reversed(i)) for i in triang]

    nan = np.zeros(ndm)*np.nan
    coords = np.array(coords)
    if "tran" in config["outline"]:
        tri_points = np.array([
            coords[idx]  if (j+1)%3 else nan
            for j,idx in enumerate(np.array(triang).reshape(-1))
        ])
    elif "long" in config["outline"]:
        tri_points = np.array([
            coords[i]  if j%2 else nan
            for j,idx in enumerate(np.array(triang)) for i in idx[IDX[j%2]] if j not in no_outline
        ])
    else:
        return

    canvas.plot_lines(tri_points,
                      style=config["line_style"]
    )

class so3:
    @classmethod
    def exp(cls, vect):
        return Rotation.from_rotvec(vect).as_matrix()

def _add_moment(artist, loc, axis):
    import meshio
    mesh_data = meshio.read(veux.assets/'chrystals_moment.stl')
    coords = mesh_data.points

    coords = np.einsum('ik, kj -> ij',  coords,
                       so3.exp([0, 0, -np.pi/4])@so3.exp(axis))
    coords = 1e-3*coords + loc
    for i in mesh_data.cells:
        if i.type == "triangle":
            triangles =  i.data #mesh_data.cells['triangle']
            break

    artist.canvas.plot_mesh(coords, triangles)


def _render(sam_file, res_file=None, **opts):
    # Configuration is determined by successively layering
    # from sources with the following priorities:
    #      defaults < file configs < kwds 

    config = veux.config.Config()


    if sam_file is None:
        raise RenderError("Expected positional argument <sam-file>")

    # Read and clean model
    if not isinstance(sam_file, dict):
        model = read_model(sam_file)
    else:
        model = sam_file

    if "RendererConfiguration" in model:
        veux.apply_config(model["RendererConfiguration"], config)

    veux.apply_config(opts, config)

    artist = veux.FrameArtist(model, **config)

    draw_extrusions(artist.model, artist.canvas, config=opts)

    # -----------------------------------------------------------

    soln = veux.state.read_state(res_file, artist.model, **opts)
    if soln is not None:
        if "time" not in opts:
            soln = soln[soln.times[-1]]

        draw_extrusions(artist.model, artist.canvas, soln, opts)
        # -----------------------------------------------------------
        _add_moment(artist,
                    loc  = [1.0, 0.0, 0.0],
                    axis = [0, np.pi/2, 0])
        # -----------------------------------------------------------

    artist.draw()
    return artist


if __name__ == "__main__":
    import veux.parser
    config = veux.parser.parse_args(sys.argv)

    try:
        artist = _render(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])


        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            import veux.server
            server = veux.server.Server(glb=artist.canvas.to_glb(),
                                        viewer=config["viewer_config"].get("name", None))
            server.run(config["server_config"].get("port", None))

        elif hasattr(artist.canvas, "to_html"):
            import veux.server
            server = veux.server.Server(html=artist.canvas.to_html())
            server.run(config["server_config"].get("port", None))

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(e, file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=sys.argv[0]), file=sys.stderr)
        sys.exit()




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
#
import sys
from collections import defaultdict

import numpy as np

try:
    import orjson as json
except ImportError:
    import json


# Constants
_EYE3 = np.eye(3)

_OUTLINES = {
    None:      None,
    "square":  np.array([[-1., -1.],
                         [ 1., -1.],
                         [ 1.,  1.],
                         [-1.,  1.]])/10,

    "tee":     np.array([[ 6.0,  0.0],
                         [ 6.0,  4.0],
                         [-6.0,  4.0],
                         [-6.0,  0.0],
                         [-2.0,  0.0],
                         [-2.0, -8.0],
                         [ 2.0, -8.0],
                         [ 2.0,  0.0]])/10
}


def _is_frame(el):
    name = el["type"].lower()
    return     "beam"  in name \
            or "dfrm"  in name \
            or "frame" in name

def _is_truss(el):
    name = el["type"].lower()
    return "truss" in name #or "twonodelink" in name


def _is_plane(el):
    name = el["type"].lower()
    return "quad" in name or "shell" in name or "tri" in name

def _is_solid(el):
    name = el["type"].lower()
    return "brick" in name


def _orient_frame(xi, xj, angle):
    """
    Calculate the coordinate transformation vector.
    xi is the location of node I, xj node J,
    and `angle` is the rotation about the local axis in degrees

    By default local axis 2 is always in the 1-Z plane, except if the object
    is vertical and then it is parallel to the global X axis.
    The definition of the local axes follows the right-hand rule.
    """

    # The local 1 axis points from node I to node J
    dx, dy, dz = e1 = xj - xi
    # Global z
    E3 = np.array([0, 0, 1])

    # In Sap2000, if the element is vertical, the local y-axis is the same as the
    # global x-axis, and the local z-axis can be obtained by cross-multiplying
    # the local x-axis with the local y-axis.
    if dx == 0 and dy == 0:
        e2 = np.array([1, 0, 0])

    # Otherwise, the plane composed of the local x-axis and the local
    # y-axis is a vertical plane. In this
    # case, the local z-axis can be obtained by the cross product of the local
    # x-axis and the global z-axis.
    else:
        e2 = np.cross(E3, e1)

    e3 = np.cross(e1, e2)

    # Rotate the local axis using the Rodrigue rotation formula
    # convert from degrees to radians
    angle = angle / 180 * np.pi
    e3r = e3 * np.cos(angle) + np.cross(e1, e3) * np.sin(angle)
    # Finally, the normalized local z-axis is returned
    return e3r / np.linalg.norm(e3r)


def read_model(filename:str, shift=None, verbose=False)->dict:
    if isinstance(filename, str) and filename.endswith(".tcl"):
        import opensees.tcl
        try:
            with open(filename, "r") as f:
                interp = opensees.tcl.exec(f.read(), silent=True, analysis=False)
        except UnicodeDecodeError:
            with open(filename, "r", encoding="latin1") as f:
                interp = opensees.tcl.exec(f.read(), silent=True, analysis=False)
        return interp.serialize()

    elif isinstance(filename, str) and (
        filename.endswith(".s2k") or filename.endswith(".$2k") or filename.endswith(".$br")) or filename.endswith(".b2k"):
        from openbim import csi
        # import veux.reader.csi as csi
        with open(filename, "r") as f:
            model = csi.create_model(csi.load(f), verbose=verbose)
        return model.asdict()

    elif isinstance(filename, str) and filename.endswith(".vtk"):
        import meshio
        from .plane import PlaneModel
        return PlaneModel(meshio.read(filename))

    elif isinstance(filename, str) and filename.endswith(".inp"):
        pass

    try:
        with open(filename,"r") as f:
            sam = json.loads(f.read())

    except TypeError:
        sam = json.loads(filename.read())

    return sam


class Model:
    def __iter__(self):
        # this method allows: nodes, cells = Model(mesh)
        return iter((self.nodes, self.cells))

    def node_location(self, tag): ...

    def node_information(self, tag): ...

    def iter_cells_tags(self, filt=None): ...

    def cell_matches(self, tag): ...

    def cell_type(self, tag):       ... # line triangle quadrilateral 

    def cell_exterior(self, tag):   ...

    def cell_interior(self, tag):   ...

    def cell_rotation(self, tag, state): ...

    def cell_position(self, tag, state):   ...

    def cell_outline(self,  tag):   ...

    def cell_information(self, tag): ...



class FrameModel:
    def __init__(self, sam:dict, shift = None, rot=None,
                 frame_outlines=None, **kwds):

        from .frame._section import SectionGeometry
        self._frame_outlines = {}
        if frame_outlines is not None:
            for key, outline in frame_outlines.items():
                if isinstance(outline, SectionGeometry):
                    self._frame_outlines[key] = [outline]
                elif isinstance(outline, list) and all(isinstance(i, SectionGeometry) for i in outline):
                    self._frame_outlines[key] = outline
                else:
                    # list of numpy arrays
                    self._frame_outlines[key] = [SectionGeometry(i) for i in outline]


        # self._frame_outlines  = {
        #     key: [outline] if isinstance(outline, SectionGeometry) else [SectionGeometry(i) for i in outline]
        #     for key, outline in frame_outlines.items()
        # }
        self._extrude_default = SectionGeometry(_OUTLINES[kwds.get("extrude_default", "square")])
        if "extrude_outline" in kwds:
            self._extrude_outline = SectionGeometry(_OUTLINES[kwds["extrude_outline"]])
        else:
            self._extrude_outline = None
        self._extrude_scale   = kwds.get("extrude_scale",   1.0)

        ndm = 3
        R = np.eye(ndm) if rot is None else rot

        if shift is None:
            shift = np.zeros(ndm)
        else:
            shift = np.asarray(shift)

        #
        self._data = _from_opensees(sam, shift, R)# output

        self.ndm = self._data["ndm"]
        self.ndf = self._data["ndf"]

    def __getitem__(self, key):
        # TODO: Remove this method
        return self._data[key]

    def cell_nodes(self, tag=None):
        if tag is None:
            if not hasattr(self, "_cell_nodes"):
                self._cell_nodes = {k: e["nodes"] for k, e in self["assembly"].items()}
            return self._cell_nodes
        else:
            return self["assembly"][tag]["nodes"]

    def cell_indices(self, tag=None):
        if not hasattr(self, "_cell_indices"):
            self._cell_indices = {
                elem["name"]: tuple(self.node_indices(n) for n in elem["nodes"])
                for elem in self["assembly"].values()
            }

        if tag is not None:
            return self._cell_indices[tag]
        else:
            return self._cell_indices

    def cell_properties(self, tag=None):
        if tag is not None:
            return self["assembly"][tag]

    def cell_prototypes(self)->"iter":
        exclude_keys = {"type", "instances", "nodes",
                        "crd", "crdTransformation"}

        if not self["prototypes"]:
            elem_types = defaultdict(dict)

            for elem in self["assembly"].values():
                type = elem["type"]
                if type not in elem_types:
                    elem_types[type]["name"] = type
                    elem_types[type]["variants"] = []
                    elem_types[type]["instances"] = [elem["name"]]
                    elem_types[type]["properties"] = {
                        k: v for k,v in elem.items() if k not in exclude_keys
                    }
                else:
                    elem_types[type]["instances"].append(elem["name"])


            elem_types = list(elem_types.values())
        else:
            elem_types = [
                {
                    "name": f"{elem['type']}<{elem['name']}>",
                    "instances": [self["assembly"][i]["name"] for i in elem["instances"]],
                    "properties":  [
                        [str(v) for k,v in elem.items() if k not in exclude_keys]
                        #for _ in range(len(elem["instances"]))
                    ]*(len(elem["instances"])),
                    "coords": [self["assembly"][i]["crd"] for i in elem["instances"]],
                    "keys":   [k for k in elem.keys() if k not in exclude_keys]

                } for elem in self["prototypes"].get("elements", [])
            ]
        return elem_types

    def iter_node_tags(self):
        for tag in self["nodes"]:
            yield tag

    def iter_cell_tags(self):
        for tag in self["assembly"]:
            yield tag

    def node_properties(self, tag=None)->dict:
        return self["nodes"][tag]

    def node_indices(self, tag=None):
        if not hasattr(self, "_node_indices"):
            self._node_indices = {
                tag: i for i, tag in enumerate(self["nodes"])
            }
        return self._node_indices[tag]

    def node_rotation(self, tag=None, state=None):
        if state is None:
            eye = np.eye(3)
            if tag is None:
                return [eye for i in self.iter_node_tags()]
            else:
                return eye
        else:
            return state.node_array(tag, dof=state.rotation)

    def node_position(self, tag=None, state=None):

        if tag is None:
            pos = np.array([n["crd"] for n in self["nodes"].values()])
        else:
            pos = self["nodes"][tag]["crd"]

        if state is not None:
            pos = pos + state.node_array(tag, dof=state.position)

        return pos

    def cell_matches(self, tag, type=None)->bool:
        elem = self["assembly"][tag]

        if type == "frame":
            return _is_frame(elem)
        
        if type == "truss":
            return _is_truss(elem)

        if type == "plane":
            return _is_plane(elem)
        
        if type == "solid":
            return _is_solid(elem) 
        
        return False

    def cell_position(self, tag, state=None):
        if isinstance(tag, list):
            return np.array([[
                    self.node_position(node, state) for node in self["assembly"][t]["nodes"]
                ] for t in tag
            ])
        else:
            return np.array([ self.node_position(node, state)
                              for node in self["assembly"][tag]["nodes"] ])


    def cell_exterior(self, tag):
        """
        return an array of node indices that outline the cell's boundary.
        This is related to finding an Eulerian path through the element's
        connectivity graph.
        """
        type = self["assembly"][tag]["type"].lower()

        if "frm" in type or "beamcol" in type:
            return self.cell_indices(tag)

        elif ("quad" in type or \
              "shell" in type and ("q" in type) or ("mitc" in type)):
            return self.cell_indices(tag)

        elif ("tri" in type or \
              "shell" in type and ("t" in type)):
            return self.cell_indices(tag)

        elif "tetra" in type:
            indices = self.cell_indices(tag)

        elif "brick" in type or "hex" in type:
            i = self.cell_indices(tag)
            if len(i) == 8:
                return [
                        i[j] for j in (0, 1, 2, 3, 0, 4, 5, 6, 7, 4, 5, 1, 3, 6, 7, 3)
                ]
            else:
                # TODO: Currently not handling higher-order bricks
                return []

        return []

    def cell_interpolation(self, tag):
        pass

    def cell_triangles(self, tag):
        """
        """
        type = self["assembly"][tag]["type"].lower()

        if self.cell_matches(tag, "frame"):
            return []

        elif ("tri" in type or
             ("shell" in type and ("dkgt" in type))):
            return [self.cell_indices(tag)]

        elif ("quad" in type or
             ("shell" in type and ("q" in type) or ("mitc" in type))):
            nodes = self.cell_indices(tag)

            if len(nodes) == 3:
                return nodes

            if len(nodes) == 4:
                return [[nodes[0], nodes[1], nodes[2]],
                        [nodes[2], nodes[3], nodes[0]]]

        elif "brick" in type:
            nodes = self.cell_indices(tag)

            if len(nodes) == 8:
                triangles = []
                for face in ((0, 3, 2, 1), (0, 1, 5, 4), (0, 4, 7, 3),
                             (6, 7, 4, 5), (6, 2, 3, 7), (6, 5, 1, 2)):
                    triangles.extend([
                            [nodes[face[0]], nodes[face[1]], nodes[face[2]]],
                            [nodes[face[2]], nodes[face[3]], nodes[face[0]]]
                    ])
                return triangles

        return []

    def add_hook():
        pass

    def frame_section(self, tag, coord=None)->"Section":
        from .frame._section import SectionGeometry

        if self.cell_matches(tag, "truss"):
            A = self["assembly"][tag].get("A", 1.0)
            o = np.sqrt(A)*np.array([[0,-1,-1],[0,1,-1],[0,1,1],[0,-1,1]], dtype=float)
            return SectionGeometry(o)

        if not self.cell_matches(tag, "frame"):
            return None

        # Initialize frame outlines
        if self._frame_outlines is None:
            # TODO: Make this dict of list of sections
            self._frame_outlines = _get_frame_outlines(self)

        sections = []
        if self._extrude_outline is not None:
            sections = [self._extrude_outline]*2 #*self._extrude_scale

        elif tag in self._frame_outlines:
            sections = self._frame_outlines[tag]

        elif self._extrude_default is not None:
            sections = [self._extrude_default]*2 #*self._extrude_scale

        if len(sections) == 0:
            print(f"Empty sections for {tag}", file=sys.stderr)
            return

        elif len(sections) == 1:
            return sections[0]

        # Interpolate coord
        elif len(sections) >= 2:
            def interpolate(values, x):
                n = len(values) - 1
                idx = np.clip(int(x * n), 0, n - 1)  # Find the lower bound index
                t = x * n - idx  # Fractional part for interpolation
                return (1 - t) * values[idx] + t * values[idx + 1]

            if coord is None:
                coord = 0.5

            exterior = np.array([s.exterior() for s in sections])
            interior = np.array([s.interior() for s in sections])
            return SectionGeometry(
                interpolate(exterior, coord),
                interpolate(interior, coord)
            )

    def frame_orientation(self, tag, state=None):
        el  = self["assembly"][tag]
        xyz = el["crd"]

        v1  = xyz[-1] - xyz[0]
        L   = np.linalg.norm(v1)
        e1  = v1/L

        if self.ndm == 2:
            v2 = np.array([0, 1, 0])

        if "yvec" in el["trsfm"] and el["trsfm"]["yvec"] is not None:
            v2  = np.array(el["trsfm"]["yvec"])

        elif "vecInLocXZPlane" in el["trsfm"]:
            v13 =  np.atleast_1d(el["trsfm"]["vecInLocXZPlane"])
            v2  = -np.cross(e1,v13)

        else:
            v3 = _orient_frame(el["crd"][0], el["crd"][1], 0)
            v2  = -np.cross(e1,v3)

        e2 = v2 / np.linalg.norm(v2)
        v3 = np.cross(e1,e2)
        e3 = v3 / np.linalg.norm(v3)
        return np.stack([e1,e2,e3])


def _from_opensees(sam: dict, shift, R):
    # Process OpenSees JSON format
    try:
        sam = sam["StructuralAnalysisModel"]
    except KeyError:
        pass

    geom = sam.get("geometry", sam.get("assembly"))

    try:
        #coord = np.array([R@n.pop("crd") for n in geom["nodes"]], dtype=float) + shift
        coord = np.array([R@n["crd"] for n in geom["nodes"]], dtype=float) + R@shift
        ndm = 3
    except:
        coord = np.array([R@[*n["crd"], 0.0] for n in geom["nodes"]], dtype=float) + shift
        ndm = 2

    nodes = {
        n["name"]: {**n, "crd": coord[i], "idx": i}
            for i,n in enumerate(geom["nodes"])
    }

#   ndm = len(next(iter(nodes.values()))["crd"])
    ndf = next(iter(nodes.values())).get("ndf", None)

    trsfm = {}
    for t in sam.get("properties", {}).get("crdTransformations", []):
        trsfm[int(t["name"])] = {
                k: val for k,val in t.items() if k != "vecInLocXZPlane"
        }
        if ndm == 3:
            trsfm[int(t["name"])]["vecInLocXZPlane"] = R@t["vecInLocXZPlane"]


    elems =  {
        e["name"]: dict(
        **e,
        crd=np.array([nodes[n]["crd"] for n in e["nodes"]], dtype=float),
        trsfm=trsfm[int(e["crdTransformation"])]
            if "crdTransformation" in e and int(e["crdTransformation"]) in trsfm
            else dict(yvec=R@e["yvec"] if "yvec" in e else None)
        ) for e in geom["elements"]
    }

    try:
        sections = {s["name"]: s for s in sam["properties"]["sections"]}
    except:
        sections = {}

    output = dict(nodes=nodes,
                  assembly=elems,
                  sam=sam,
                  sections=sections,
                  prototypes=sam.get("prototypes", {}),
                  ndm=ndm,
                  ndf=ndf
    )

    return output


def collect_outlines(model):
    return _get_frame_outlines(_from_opensees(model, [0, 0, 0], np.eye(3)))

def _add_section_shape(section, sections, outlines):
    import scipy.spatial

    # Rotation to change coordinates from x-y to z-y
    R = np.array(((0,-1),
                  (1, 0))).T

    if "section" in section:
        # Treat aggregated sections
        if section["section"] not in outlines:
            outlines[section["name"]] = _add_section_shape(sections[section["section"]], sections, outlines)
        else:
            outlines[section["name"]] = outlines[section["section"]]

    elif "bounding_polygon" in section:
        outlines[section["name"]] = [R@s for s in section["bounding_polygon"]]

    elif "fibers" in section:
        points = np.array([f["coord"] for f in section["fibers"]])
        try:
            outlines[section["name"]] = points[scipy.spatial.ConvexHull(points).vertices]
        except scipy.spatial._qhull.QhullError as e:
            import warnings
            from veux.utility.alpha_shape import alpha_shape
            outlines[section["name"]] = alpha_shape(points, 1)
            warnings.warn(str(e))


def _get_frame_outlines(model):
    shapes = {}
    for name,section in model["sections"].items():
        _add_section_shape(section, model["sections"], shapes)

    # Function to check if list of lists is homogeneous
    homogeneous = lambda lst: (
            isinstance(lst, list) and \
              all(isinstance(x, list) and \
                  len(set(map(len, lst))) == 1 and all(isinstance(xx, list) and len(set(map(len, x))) == 1 for xx in x) for x in lst
              )
    )

    from shps.frame.extrude import Section
    outlines = {}
    for elem in model["assembly"].values():
        if "sections" in elem:
            elem_shapes = [shapes[i] for i in elem["sections"] if i in shapes]
            if len(elem_shapes) == 0:
                continue
            elif not homogeneous(elem_shapes):
                elem_shapes = np.array(elem_shapes[0])
            else:
                elem_shapes = np.array(elem_shapes)

            outlines[elem["name"]] = [Section(shape) for shape in elem_shapes]

    return outlines

